@@ -1,7 +1,7 @@
 /* cgmanager
  *
- * Copyright © 2013 Stphane Graber
- * Author: Stphane Graber <stgraber@ubuntu.com>
+ * Copyright © 2013 Stephane Graber
+ * Author: Stephane Graber <stgraber@ubuntu.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2, as
@@ -18,13 +18,14 @@
  */
 
 #include <frontend.h>
+#include <sys/resource.h>
 
 /*
  * Maximum depth of directories we allow in Create
  * Default is 16.  Figure 4 directories per level of container
  * nesting (/user/1000.user/c2.session/c1), that lets us nest
  * 4 containers deep.
- * */
+ */
 static int maxdepth = 16;
 
 /* GetPidCgroup */
@@ -35,13 +36,15 @@ int get_pid_cgroup_main(void *parent, const char *controller,struct ucred p,
 
 	// Get r's current cgroup in rcgpath
 	if (!compute_pid_cgroup(r.pid, controller, "", rcgpath, NULL)) {
-		nih_error("%s: Could not determine the requestor cgroup", __func__);
+		nih_error("%s: Could not determine the requestor's cgroup for %s",
+				__func__, controller);
 		return -1;
 	}
 
 	// Get v's cgroup in vcgpath
 	if (!compute_pid_cgroup(v.pid, controller, "", vcgpath, NULL)) {
-		nih_error("%s: Could not determine the victim cgroup", __func__);
+		nih_error("%s: Could not determine the victim cgroup for %s",
+				__func__, controller);
 		return -1;
 	}
 
@@ -60,13 +63,49 @@ int get_pid_cgroup_main(void *parent, const char *controller,struct ucred p,
 	return 0;
 }
 
+/* GetPidCgroupAbs */
+int get_pid_cgroup_abs_main(void *parent, const char *controller,struct ucred p,
+			 struct ucred r, struct ucred v, char **output)
+{
+	char rcgpath[MAXPATHLEN], vcgpath[MAXPATHLEN];
+
+	// Get p's current cgroup in rcgpath
+	if (!compute_pid_cgroup(p.pid, controller, "", rcgpath, NULL)) {
+		nih_error("%s: Could not determine the requestor's cgroup for %s",
+				__func__, controller);
+		return -1;
+	}
+
+	// Get v's cgroup in vcgpath
+	if (!compute_pid_cgroup(v.pid, controller, "", vcgpath, NULL)) {
+		nih_error("%s: Could not determine the victim cgroup for %s",
+				__func__, controller);
+		return -1;
+	}
+
+	// Make sure v's cgroup is under p's
+	int rlen = strlen(rcgpath);
+	if (strncmp(rcgpath, vcgpath, rlen) != 0) {
+		nih_error("%s: v (%d)'s cgroup is not below p (%d)'s", __func__,
+			v.pid, p.pid);
+		return -1;
+	}
+	if (strlen(vcgpath) == rlen)
+		*output = NIH_MUST (nih_strdup(parent, "/") );
+	else
+		*output = NIH_MUST (nih_strdup(parent, vcgpath + rlen) );
+
+	return 0;
+}
+
 static bool victim_under_proxy_cgroup(char *rcgpath, pid_t v,
 		const char *controller)
 {
 	char vcgpath[MAXPATHLEN];
 
 	if (!compute_pid_cgroup(v, controller, "", vcgpath, NULL)) {
-		nih_error("%s: Could not determine the victim's cgroup", __func__);
+		nih_error("%s: Could not determine the victim's cgroup for %s",
+				__func__, controller);
 		return false;
 	}
 	if (strncmp(vcgpath, rcgpath, strlen(rcgpath)) != 0)
@@ -74,29 +113,19 @@ static bool victim_under_proxy_cgroup(char *rcgpath, pid_t v,
 	return true;
 }
 
-int do_move_pid_main(const char *controller, const char *cgroup, struct ucred p,
+int per_ctrl_move_pid_main(const char *controller, const char *cgroup, struct ucred p,
 		struct ucred r, struct ucred v, bool escape)
 {
 	char rcgpath[MAXPATHLEN], path[MAXPATHLEN];
 	FILE *f;
 	pid_t query = r.pid;
 
-	if (!sane_cgroup(cgroup)) {
-		nih_error("%s: unsafe cgroup", __func__);
-		return -1;
-	}
-
-	// verify that ucred.pid may move target pid
-	if (!may_move_pid(r.pid, r.uid, v.pid)) {
-		nih_error("%s: %d may not move %d", __func__, r.pid, v.pid);
-		return -1;
-	}
-
 	// Get r's current cgroup in rcgpath
 	if (escape)
 		query = p.pid;
 	if (!compute_pid_cgroup(query, controller, "", rcgpath, NULL)) {
-		nih_error("%s: Could not determine the requested cgroup", __func__);
+		nih_error("%s: Could not determine the requestor's cgroup for %s",
+                __func__, controller);
 		return -1;
 	}
 
@@ -150,6 +179,49 @@ int do_move_pid_main(const char *controller, const char *cgroup, struct ucred p,
 	return 0;
 }
 
+int do_move_pid_main(const char *controller, const char *cgroup, struct ucred p,
+		struct ucred r, struct ucred v, bool escape)
+{
+	nih_local char *c = NULL;
+	char *tok;
+	int ret;
+
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
+		return -1;
+	}
+
+	// verify that ucred.pid may move target pid
+	if (!may_move_pid(r.pid, r.uid, v.pid)) {
+		nih_error("%s: %d may not move %d", __func__, r.pid, v.pid);
+		return -1;
+	}
+
+	if (strcmp(controller, "all") != 0 && !strchr(controller, ','))
+		return per_ctrl_move_pid_main(controller, cgroup, p, r, v, escape);
+
+	if (strcmp(controller, "all") == 0) {
+		if (!all_controllers)
+			return 0;
+		c = NIH_MUST( nih_strdup(NULL, all_controllers) );
+	} else {
+		c = NIH_MUST( nih_strdup(NULL, controller) );
+		do_prune_comounts(c);
+	}
+	tok = strtok(c, ",");
+	while (tok) {
+		ret = per_ctrl_move_pid_main(tok, cgroup, p, r, v, escape);
+		if (ret == -2)  // permission denied - ignore for group requests
+			goto next;
+		if (ret != 0)
+			return -1;
+next:
+		tok = strtok(NULL, ",");
+	}
+
+	return 0;
+}
+
 int move_pid_main(const char *controller, const char *cgroup, struct ucred p,
 		struct ucred r, struct ucred v)
 {
@@ -168,7 +240,7 @@ int move_pid_abs_main(const char *controller, const char *cgroup, struct ucred p
 	return do_move_pid_main(controller, cgroup, p, r, v, true);
 }
 
-int create_main(const char *controller, const char *cgroup, struct ucred p,
+int do_create_main(const char *controller, const char *cgroup, struct ucred p,
 		struct ucred r, int32_t *existed)
 {
 	int ret, depth;
@@ -178,17 +250,10 @@ int create_main(const char *controller, const char *cgroup, struct ucred p,
 	char *p1, *p2, oldp2;
 
 	*existed = 1;
-	if (!cgroup || ! *cgroup)  // nothing to do
-		return 0;
-
-	if (!sane_cgroup(cgroup)) {
-		nih_error("%s: unsafe cgroup", __func__);
-		return -1;
-	}
-
 	// Get r's current cgroup in rcgpath
 	if (!compute_pid_cgroup(r.pid, controller, "", rcgpath, &depth)) {
-		nih_error("%s: Could not determine the requested cgroup", __func__);
+		nih_error("%s: Could not determine the requestor's cgroup for %s",
+                __func__, controller);
 		return -1;
 	}
 
@@ -224,14 +289,14 @@ int create_main(const char *controller, const char *cgroup, struct ucred p,
 			if (!may_access(r.pid, r.uid, r.gid, path, O_RDONLY)) {
 				nih_error("%s: pid %d (uid %u gid %u) may not look under %s", __func__,
 					r.pid, r.uid, r.gid, path);
-				return -1;
+				return -2;
 			}
 			goto next;
 		}
 		if (!may_access(r.pid, r.uid, r.gid, dirpath, O_RDWR)) {
 			nih_error("%s: pid %d (uid %u gid %u) may not create under %s", __func__,
 				r.pid, r.uid, r.gid, dirpath);
-			return -1;
+			return -2;
 		}
 		ret = mkdir(path, 0755);
 		if (ret < 0) {  // Should we ignore EEXIST?  Ok, but don't chown.
@@ -240,7 +305,7 @@ int create_main(const char *controller, const char *cgroup, struct ucred p,
 				goto next;
 			}
 			nih_error("%s: failed to create %s", __func__, path);
-			return -1;
+			return -2;
 		}
 		if (!chown_cgroup_path(path, r.uid, r.gid, true)) {
 			nih_error("%s: Failed to change ownership on %s to %u:%u", __func__,
@@ -263,30 +328,60 @@ next:
 	return 0;
 }
 
-int chown_main(const char *controller, const char *cgroup, struct ucred p,
-		struct ucred r, struct ucred v)
+int create_main(const char *controller, const char *cgroup, struct ucred p,
+		struct ucred r, int32_t *existed)
 {
-	char rcgpath[MAXPATHLEN];
-	nih_local char *path = NULL;
-	uid_t uid;
+	nih_local char *c = NULL;
+	char *tok;
+	int ret;
 
-	/* If caller is not root in his userns, then he can't chown, as
-	 * that requires privilege over two uids */
-	if (r.uid) {
-		if (!hostuid_to_ns(r.uid, r.pid, &uid) || uid != 0) {
-			nih_error("%s: Chown requested by non-root uid %u", __func__, r.uid);
-			return -1;
-		}
-	}
+	*existed = 1;
+	if (!cgroup || ! *cgroup)  // nothing to do
+		return 0;
 
 	if (!sane_cgroup(cgroup)) {
 		nih_error("%s: unsafe cgroup", __func__);
 		return -1;
 	}
 
+	if (strcmp(controller, "all") != 0 && !strchr(controller, ','))
+		return do_create_main(controller, cgroup, p, r, existed);
+
+	if (strcmp(controller, "all") == 0) {
+		if (!all_controllers)
+			return 0;
+		c = NIH_MUST( nih_strdup(NULL, all_controllers) );
+	} else {
+		c = NIH_MUST( nih_strdup(NULL, controller) );
+		do_prune_comounts(c);
+	}
+	tok = strtok(c, ",");
+	while (tok) {
+		int32_t e = 1;
+		ret = do_create_main(tok, cgroup, p, r, &e);
+		if (ret == -2)  // permission denied - ignore for group requests
+			goto next;
+		if (ret != 0)
+			return -1;
+		if (!e)
+			*existed = 0;
+next:
+		tok = strtok(NULL, ",");
+	}
+
+	return 0;
+}
+
+int do_chown_main(const char *controller, const char *cgroup, struct ucred p,
+		struct ucred r, struct ucred v)
+{
+	char rcgpath[MAXPATHLEN];
+	nih_local char *path = NULL;
+
 	// Get r's current cgroup in rcgpath
 	if (!compute_pid_cgroup(r.pid, controller, "", rcgpath, NULL)) {
-		nih_error("%s: Could not determine the requested cgroup", __func__);
+		nih_error("%s: Could not determine the requestor's cgroup for %s",
+                __func__, controller);
 		return -1;
 	}
 	/* rcgpath + / + cgroup + \0 */
@@ -303,45 +398,82 @@ int chown_main(const char *controller, const char *cgroup, struct ucred p,
 	if (!may_access(r.pid, r.uid, r.gid, path, O_RDONLY)) {
 		nih_error("%s: pid %d (uid %u gid %u) may not read under %s", __func__,
 			r.pid, r.uid, r.gid, path);
-		return -1;
+		return -2;
 	}
 
 	// does r have privilege over the cgroup dir?
 	if (!may_access(r.pid, r.uid, r.gid, path, O_RDWR)) {
 		nih_error("%s: Pid %d may not chown %s\n", __func__, r.pid, path);
-		return -1;
+		return -2;
 	}
 
 	// go ahead and chown it.
 	if (!chown_cgroup_path(path, v.uid, v.gid, false)) {
 		nih_error("%s: Failed to change ownership on %s to %u:%u", __func__,
 			path, v.uid, v.gid);
-		return -1;
+		return -2;
 	}
 
 	return 0;
 }
 
-int chmod_main(const char *controller, const char *cgroup, const char *file,
-		struct ucred p, struct ucred r, int mode)
+int chown_main(const char *controller, const char *cgroup, struct ucred p,
+		struct ucred r, struct ucred v)
 {
-	char rcgpath[MAXPATHLEN];
-	nih_local char *path = NULL;
+	uid_t uid;
+	nih_local char *c = NULL;
+	char *tok;
+	int ret;
+
+	/* If caller is not root in his userns, then he can't chown, as
+	 * that requires privilege over two uids */
+	if (r.uid) {
+		if (!hostuid_to_ns(r.uid, r.pid, &uid) || uid != 0) {
+			nih_error("%s: Chown requested by non-root uid %u", __func__, r.uid);
+			return -1;
+		}
+	}
 
 	if (!sane_cgroup(cgroup)) {
 		nih_error("%s: unsafe cgroup", __func__);
 		return -1;
 	}
 
-	if (file && ( strchr(file, '/') || strchr(file, '\\')) ) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-				"invalid file");
-		return -1;
+	if (strcmp(controller, "all") != 0 && !strchr(controller, ','))
+		return do_chown_main(controller, cgroup, p, r, v);
+
+	if (strcmp(controller, "all") == 0) {
+		if (!all_controllers)
+			return 0;
+		c = NIH_MUST( nih_strdup(NULL, all_controllers) );
+	} else {
+		c = NIH_MUST( nih_strdup(NULL, controller) );
+		do_prune_comounts(c);
 	}
+	tok = strtok(c, ",");
+	while (tok) {
+		ret = do_chown_main(tok, cgroup, p, r, v);
+		if (ret == -2)  // permission denied - ignore for group requests
+			goto next;
+		if (ret != 0)
+			return -1;
+next:
+		tok = strtok(NULL, ",");
+	}
+
+	return 0;
+}
+
+int do_chmod_main(const char *controller, const char *cgroup, const char *file,
+		struct ucred p, struct ucred r, int mode)
+{
+	char rcgpath[MAXPATHLEN];
+	nih_local char *path = NULL;
 
 	// Get r's current cgroup in rcgpath
 	if (!compute_pid_cgroup(r.pid, controller, "", rcgpath, NULL)) {
-		nih_error("%s: Could not determine the requested cgroup", __func__);
+		nih_error("%s: Could not determine the requestor's cgroup for %s",
+                __func__, controller);
 		return -1;
 	}
 
@@ -356,22 +488,65 @@ int chmod_main(const char *controller, const char *cgroup, const char *file,
 	if (!may_access(r.pid, r.uid, r.gid, path, O_RDONLY)) {
 		nih_error("%s: pid %d (uid %u gid %u) may not read under %s", __func__,
 			r.pid, r.uid, r.gid, path);
-		return -1;
+		return -2;
 	}
 
 	// does r have privilege over the cgroup dir?
 	if (!may_access(r.pid, r.uid, r.gid, path, O_RDWR)) {
 		nih_error("%s: Pid %d may not chmod %s\n", __func__, r.pid, path);
-		return -1;
+		return -2;
 	}
 
 	// go ahead and chmod it.
 	if (!chmod_cgroup_path(path, mode)) {
 		nih_error("%s: Failed to change mode on %s to %d", __func__,
 			path, mode);
+		return -2;
+	}
+
+	return 0;
+}
+
+int chmod_main(const char *controller, const char *cgroup, const char *file,
+		struct ucred p, struct ucred r, int mode)
+{
+	nih_local char *c = NULL;
+	char *tok;
+	int ret;
+
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
 		return -1;
 	}
 
+	if (file && ( strchr(file, '/') || strchr(file, '\\')) ) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+				"invalid file");
+		return -1;
+	}
+
+	if (strcmp(controller, "all") != 0 && !strchr(controller, ','))
+		return do_chmod_main(controller, cgroup, file, p, r, mode);
+
+	if (strcmp(controller, "all") == 0) {
+		if (!all_controllers)
+			return 0;
+		c = NIH_MUST( nih_strdup(NULL, all_controllers) );
+	} else {
+		c = NIH_MUST( nih_strdup(NULL, controller) );
+		do_prune_comounts(c);
+	}
+	tok = strtok(c, ",");
+	while (tok) {
+		ret = do_chmod_main(tok, cgroup, file, p, r, mode);
+		if (ret == -2)  // permission denied - ignore for group requests
+			goto next;
+		if (ret != 0)
+			return -1;
+next:
+		tok = strtok(NULL, ",");
+	}
+
 	return 0;
 }
 
@@ -386,7 +561,8 @@ int get_value_main(void *parent, const char *controller, const char *cgroup,
 	}
 
 	if (!compute_pid_cgroup(r.pid, controller, cgroup, path, NULL)) {
-		nih_error("%s: Could not determine the requested cgroup", __func__);
+		nih_error("%s: Could not determine the requested cgroup (%s:%s)",
+                __func__, controller, cgroup);
 		return -1;
 	}
 
@@ -435,7 +611,8 @@ int set_value_main(const char *controller, const char *cgroup,
 	}
 
 	if (!compute_pid_cgroup(r.pid, controller, cgroup, path, NULL)) {
-		nih_error("%s: Could not determine the requested cgroup", __func__);
+		nih_error("%s: Could not determine the requested cgroup (%s:%s)",
+                __func__, controller, cgroup);
 		return -1;
 	}
 
@@ -500,7 +677,6 @@ static int recursive_rmdir(char *path)
 {
 	struct dirent dirent, *direntp;
 	DIR *dir;
-	char pathname[MAXPATHLEN];
 	int failed = 0;
 
 	dir = opendir(path);
@@ -511,6 +687,7 @@ static int recursive_rmdir(char *path)
 
 	while (!readdir_r(dir, &dirent, &direntp)) {
 		struct stat mystat;
+		nih_local char *pathname = NULL;
 		int rc;
 
 		if (!direntp)
@@ -518,11 +695,7 @@ static int recursive_rmdir(char *path)
 		if (!strcmp(direntp->d_name, ".") ||
 		    !strcmp(direntp->d_name, ".."))
 			continue;
-		rc = snprintf(pathname, MAXPATHLEN, "%s/%s", path, direntp->d_name);
-		if (rc < 0 || rc >= MAXPATHLEN) {
-			failed = 1;
-			continue;
-		}
+		pathname = NIH_MUST( nih_sprintf(NULL, "%s/%s", path, direntp->d_name) );
 		rc = lstat(pathname, &mystat);
 		if (rc) {
 			failed = 1;
@@ -542,7 +715,7 @@ static int recursive_rmdir(char *path)
 	return failed ? -1 : 0;
 }
 
-int remove_main(const char *controller, const char *cgroup, struct ucred p,
+int do_remove_main(const char *controller, const char *cgroup, struct ucred p,
 		struct ucred r, int recursive, int32_t *existed)
 {
 	char rcgpath[MAXPATHLEN];
@@ -551,15 +724,10 @@ int remove_main(const char *controller, const char *cgroup, struct ucred p,
 	char *p1;
 
 	*existed = 1;
-
-	if (!sane_cgroup(cgroup)) {
-		nih_error("%s: unsafe cgroup", __func__);
-		return -1;
-	}
-
 	// Get r's current cgroup in rcgpath
 	if (!compute_pid_cgroup(r.pid, controller, "", rcgpath, NULL)) {
-		nih_error("%s: Could not determine the requested cgroup", __func__);
+		nih_error("%s: Could not determine the requestor's cgroup for %s",
+                __func__, controller);
 		return -1;
 	}
 
@@ -590,13 +758,13 @@ int remove_main(const char *controller, const char *cgroup, struct ucred p,
 	if (!may_access(r.pid, r.uid, r.gid, copy, O_WRONLY)) {
 		nih_error("%s: pid %d (%u:%u) may not remove %s", __func__,
 			r.pid, r.uid, r.gid, copy);
-		return -1;
+		return -2;
 	}
 
 	if (!recursive) {
 		if (rmdir(working) < 0) {
 			nih_error("%s: Failed to remove %s: %s", __func__, working, strerror(errno));
-			return -1;
+			return errno == EPERM ? -2 : -1;
 		}
 	} else if (recursive_rmdir(working) < 0)
 			return -1;
@@ -606,11 +774,53 @@ int remove_main(const char *controller, const char *cgroup, struct ucred p,
 	return 0;
 }
 
+int remove_main(const char *controller, const char *cgroup, struct ucred p,
+		struct ucred r, int recursive, int32_t *existed)
+{
+	nih_local char *c = NULL;
+	char *tok;
+	int ret;
+
+	*existed = 1;
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
+		return -1;
+	}
+
+	if (strcmp(controller, "all") != 0 && !strchr(controller, ','))
+		return do_remove_main(controller, cgroup, p, r, recursive, existed);
+
+	if (strcmp(controller, "all") == 0) {
+		if (!all_controllers)
+			return 0;
+		c = NIH_MUST( nih_strdup(NULL, all_controllers) );
+	} else {
+		c = NIH_MUST( nih_strdup(NULL, controller) );
+		do_prune_comounts(c);
+	}
+	tok = strtok(c, ",");
+	while (tok) {
+		int32_t e = 1;
+		ret = do_remove_main(tok, cgroup, p, r, recursive, &e);
+		if (ret == -2)  // permission denied - ignore for group requests
+			goto next;
+		if (ret != 0)
+			return -1;
+		if (!e)
+			*existed = 0;
+next:
+		tok = strtok(NULL, ",");
+	}
+
+	return 0;
+}
+
 int get_tasks_main(void *parent, const char *controller, const char *cgroup,
 			struct ucred p, struct ucred r, int32_t **pids)
 {
 	char path[MAXPATHLEN];
 	const char *key = "tasks";
+	int alloced_pids = 0, nrpids = 0;
 
 	if (!sane_cgroup(cgroup)) {
 		nih_error("%s: unsafe cgroup", __func__);
@@ -618,7 +828,8 @@ int get_tasks_main(void *parent, const char *controller, const char *cgroup,
 	}
 
 	if (!compute_pid_cgroup(r.pid, controller, cgroup, path, NULL)) {
-		nih_error("%s: Could not determine the requested cgroup", __func__);
+		nih_error("%s: Could not determine the requested cgroup (%s:%s)",
+                __func__, controller, cgroup);
 		return -1;
 	}
 
@@ -637,7 +848,136 @@ int get_tasks_main(void *parent, const char *controller, const char *cgroup,
 	strncat(path, "/", MAXPATHLEN-1);
 	strncat(path, key, MAXPATHLEN-1);
 
-	return file_read_pids(parent, path, pids);
+	*pids = NULL;
+	if (file_read_pids(parent, path, pids, &alloced_pids, &nrpids) < 0) {
+		nih_free(*pids);
+		return -1;
+	}
+	return nrpids;
+}
+
+int do_collect_tasks(void *parent, char *path, int32_t **pids,
+			int *alloced_pids, int *nrpids)
+{
+	struct dirent dirent, *direntp;
+	DIR *dir;
+	const char *key = "tasks";
+
+	dir = opendir(path);
+	if (!dir) {
+		nih_warn("%s: Failed to open dir %s for recursive deletion", __func__, path);
+		return -2;
+	}
+
+	while (!readdir_r(dir, &dirent, &direntp)) {
+		struct stat mystat;
+		nih_local char *childname = NULL;
+		int rc;
+
+		if (!direntp)
+			break;
+		if (!strcmp(direntp->d_name, ".") ||
+		    !strcmp(direntp->d_name, ".."))
+			continue;
+		childname = NIH_MUST( nih_sprintf(NULL, "%s/%s", path, direntp->d_name) );
+		rc = lstat(childname, &mystat);
+		if (rc)
+			continue;
+		if (S_ISDIR(mystat.st_mode))
+			if (do_collect_tasks(parent, childname, pids,
+						alloced_pids, nrpids) == -1)
+				nih_info("%s: error descending subdirs", __func__);
+	}
+
+	closedir(dir);
+
+	/* Get tasks for this directory */
+
+	/* append the filename */
+	if (strlen(path) + strlen(key) + 2 > MAXPATHLEN) {
+		nih_error("%s: filename too long for cgroup %s key %s", __func__, path, key);
+		return -1;
+	}
+
+	strncat(path, "/", MAXPATHLEN-1);
+	strncat(path, key, MAXPATHLEN-1);
+
+	return file_read_pids(parent, path, pids, alloced_pids, nrpids);
+}
+
+int collect_tasks(void *parent, const char *controller, const char *cgroup,
+		struct ucred p, struct ucred r, int32_t **pids,
+		int *alloced_pids, int *nrpids)
+{
+	char path[MAXPATHLEN];
+
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
+		return -1;
+	}
+
+	if (!compute_pid_cgroup(r.pid, controller, cgroup, path, NULL)) {
+		nih_error("%s: Could not determine the requested cgroup (%s:%s)",
+                __func__, controller, cgroup);
+		return -2;
+	}
+
+	/* Check access rights to the cgroup directory */
+	if (!may_access(r.pid, r.uid, r.gid, path, O_RDONLY)) {
+		nih_error("%s: Pid %d may not access %s\n", __func__, r.pid, path);
+		return -2;
+	}
+
+	return do_collect_tasks(parent, path, pids, alloced_pids, nrpids);
+}
+
+int get_tasks_recursive_main(void *parent, const char *controller,
+		const char *cgroup, struct ucred p, struct ucred r, int32_t **pids)
+{
+	nih_local char *c = NULL;
+	char *tok;
+	int ret;
+	int alloced_pids = 0, nrpids = 0;
+
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
+		return -1;
+	}
+
+	*pids = NULL;
+
+	if (strcmp(controller, "all") != 0 && !strchr(controller, ',')) {
+		if (collect_tasks(parent, controller, cgroup, p, r, pids,
+				&alloced_pids, &nrpids) < 0) {
+			nih_free(*pids);
+			return -1;
+		}
+		return nrpids;
+	}
+
+	if (strcmp(controller, "all") == 0) {
+		if (!all_controllers)
+			return 0;
+		c = NIH_MUST( nih_strdup(NULL, all_controllers) );
+	} else {
+		c = NIH_MUST( nih_strdup(NULL, controller) );
+		do_prune_comounts(c);
+	}
+	tok = strtok(c, ",");
+	while (tok) {
+		ret = collect_tasks(parent, tok, cgroup, p, r, pids,
+				&alloced_pids, &nrpids);
+		if (ret == -2)  // permission denied - ignore
+			goto next;
+		if (ret != 0) {
+			nih_free(*pids);
+			return -1;
+		}
+next:
+		tok = strtok(NULL, ",");
+	}
+
+	return nrpids;
 }
 
 int list_children_main(void *parent, const char *controller, const char *cgroup,
@@ -652,7 +992,8 @@ int list_children_main(void *parent, const char *controller, const char *cgroup,
 	}
 
 	if (!compute_pid_cgroup(r.pid, controller, cgroup, path, NULL)) {
-		nih_error("%s: Could not determine the requested cgroup", __func__);
+		nih_error("%s: Could not determine the requested cgroup (%s:%s)",
+                __func__, controller, cgroup);
 		return -1;
 	}
 
@@ -662,24 +1003,25 @@ int list_children_main(void *parent, const char *controller, const char *cgroup,
 		return -1;
 	}
 
-	return get_child_directories(parent, path, output);
+	return get_directory_children(parent, path, output, DT_DIR);
 }
 
-int remove_on_empty_main(const char *controller, const char *cgroup,
+int do_remove_on_empty_main(const char *controller, const char *cgroup,
 		struct ucred p, struct ucred r)
 {
 	char rcgpath[MAXPATHLEN];
 	size_t cgroup_len;
 	nih_local char *working = NULL, *wcgroup = NULL;
 
-	if (!sane_cgroup(cgroup)) {
-		nih_error("%s: unsafe cgroup", __func__);
-		return -1;
+	if (was_premounted(controller)) {
+		nih_error("remove-on-empty request for pre-mounted controller");
+		return -2;
 	}
 
 	// Get r's current cgroup in rcgpath
 	if (!compute_pid_cgroup(r.pid, controller, "", rcgpath, NULL)) {
-		nih_error("%s: Could not determine the requested cgroup", __func__);
+		nih_error("%s: Could not determine the requestor's cgroup for %s",
+                __func__, controller);
 		return -1;
 	}
 
@@ -717,6 +1059,175 @@ int remove_on_empty_main(const char *controller, const char *cgroup,
 
 	return 0;
 }
+
+int remove_on_empty_main(const char *controller, const char *cgroup,
+		struct ucred p, struct ucred r)
+{
+	nih_local char *c = NULL;
+	char *tok;
+	int ret;
+
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
+		return -1;
+	}
+
+	if (strcmp(controller, "all") != 0 && !strchr(controller, ','))
+		return do_remove_on_empty_main(controller, cgroup, p, r);
+
+	if (strcmp(controller, "all") == 0) {
+		if (!all_controllers)
+			return 0;
+		c = NIH_MUST( nih_strdup(NULL, all_controllers) );
+	} else {
+		c = NIH_MUST( nih_strdup(NULL, controller) );
+		do_prune_comounts(c);
+	}
+	tok = strtok(c, ",");
+	while (tok) {
+		ret = do_remove_on_empty_main(tok, cgroup, p, r);
+		if (ret == -2)  // pre-mounted, autoremove not an option, ignore
+			goto next;
+		if (ret != 0)
+			return -1;
+next:
+		tok = strtok(NULL, ",");
+	}
+
+	return 0;
+}
+
+void do_recursive_prune(char *path, bool autoremove)
+{
+	nih_local char *releasefile = NULL;
+	struct dirent dirent, *direntp;
+	DIR *dir;
+
+	if (!autoremove)
+		goto remove;
+
+	releasefile = NIH_MUST( nih_strdup(NULL, path) );
+	NIH_MUST( nih_strcat(&releasefile, NULL, "/notify_on_release") );
+
+	if (!set_value_trusted(releasefile, "1\n"))
+		nih_info("Failed to set remove-on-empty for %s\n", path);
+
+remove:
+	dir = opendir(path);
+	if (!dir) {
+		nih_warn("%s: Failed to open dir %s for recursive deletion", __func__, path);
+		return;
+	}
+
+	while (!readdir_r(dir, &dirent, &direntp)) {
+		struct stat mystat;
+		int rc;
+		nih_local char *pathname = NULL;
+
+		if (!direntp)
+			break;
+		if (!strcmp(direntp->d_name, ".") ||
+		    !strcmp(direntp->d_name, ".."))
+			continue;
+		pathname = NIH_MUST( nih_sprintf(NULL, "%s/%s", path, direntp->d_name) );
+		rc = lstat(pathname, &mystat);
+		if (rc)
+			continue;
+		if (S_ISDIR(mystat.st_mode))
+			do_recursive_prune(pathname, autoremove);
+	}
+
+	closedir(dir);
+	rmdir(path);
+}
+
+int do_prune_main(const char *controller, const char *cgroup,
+		struct ucred p, struct ucred r)
+{
+	char rcgpath[MAXPATHLEN];
+	size_t cgroup_len;
+	nih_local char *working = NULL, *wcgroup = NULL;
+
+	// Get r's current cgroup in rcgpath
+	if (!compute_pid_cgroup(r.pid, controller, "", rcgpath, NULL)) {
+		nih_error("%s: Could not determine the requestor's cgroup for %s",
+                __func__, controller);
+		return -1;
+	}
+
+	cgroup_len = strlen(cgroup);
+
+	if (strlen(rcgpath) + cgroup_len > MAXPATHLEN) {
+		nih_error("%s: Path name too long", __func__);
+		return -1;
+	}
+
+	wcgroup = NIH_MUST( nih_strdup(NULL, cgroup) );
+	if (!normalize_path(wcgroup))
+		return -1;
+
+	working = NIH_MUST( nih_strdup(NULL, rcgpath) );
+	NIH_MUST( nih_strcat(&working, NULL, "/") );
+	NIH_MUST( nih_strcat(&working, NULL, wcgroup) );
+
+	if (!dir_exists(working)) {
+		return -1;
+	}
+	// must have write access
+	if (!may_access(r.pid, r.uid, r.gid, working, O_WRONLY)) {
+		nih_error("%s: pid %d (%u:%u) may not remove %s", __func__,
+			r.pid, r.uid, r.gid, working);
+		return -1;
+	}
+
+	do_recursive_prune(working, !was_premounted(controller));
+
+	return 0;
+}
+
+int prune_main(const char *controller, const char *cgroup,
+		struct ucred p, struct ucred r)
+{
+	nih_local char *c = NULL;
+	char *tok;
+	int ret;
+
+	if (!sane_cgroup(cgroup)) {
+		nih_error("%s: unsafe cgroup", __func__);
+		return -1;
+	}
+
+	if (strcmp(controller, "all") != 0 && !strchr(controller, ','))
+		return do_prune_main(controller, cgroup, p, r);
+
+	if (strcmp(controller, "all") == 0) {
+		if (!all_controllers)
+			return 0;
+		c = NIH_MUST( nih_strdup(NULL, all_controllers) );
+	} else {
+		c = NIH_MUST( nih_strdup(NULL, controller) );
+		do_prune_comounts(c);
+	}
+	tok = strtok(c, ",");
+	while (tok) {
+		ret = do_prune_main(tok, cgroup, p, r);
+		if (ret != 0)
+			nih_warn("do_prune_main for %s: %s failed", tok, cgroup);
+		tok = strtok(NULL, ",");
+	}
+
+	return 0;
+}
+
+int list_controllers_main(void *parent, char ***output)
+{
+	*output = NULL;
+
+	do_list_controllers(parent, output);
+
+	return 0;
+}
+
 char *extra_cgroup_mounts;
 
 static int
@@ -739,6 +1250,8 @@ static NihOption options[] = {
 		NULL, "subsystems to mount", NULL, my_setter },
 	{ 0, "daemon", N_("Detach and run in the background"),
 		NULL, NULL, &daemonise, NULL },
+	{ 0, "sigstop", N_("Raise SIGSTOP when ready"),
+		NULL, NULL, &sigstop, NULL },
 
 	NIH_OPTION_LAST
 };
@@ -814,6 +1327,7 @@ main (int argc, char *argv[])
 	int		ret;
 	DBusServer *	server;
 	struct stat sb;
+	struct rlimit newrlimit;
 
 	nih_main_init (argv[0]);
 
@@ -870,6 +1384,12 @@ main (int argc, char *argv[])
 		setns_user_supported = true;
 	}
 
+	newrlimit.rlim_cur = 10000;
+	newrlimit.rlim_max = 10000;
+	if (setrlimit(RLIMIT_NOFILE, &newrlimit) < 0)
+		nih_warn("Failed to increase open file limit: %s",
+			strerror(errno));
+
 	/* Become daemon */
 	if (daemonise) {
 		if (nih_main_daemonise () < 0) {
@@ -884,6 +1404,9 @@ main (int argc, char *argv[])
 		}
 	}
 
+	if (sigstop)
+		raise(SIGSTOP);
+
 	ret = nih_main_loop ();
 
 	/* Destroy any PID file we may have created */
