@@ -1,8 +1,8 @@
-/* frontend.h: definitions of the dbus and scm-enhanced dbus
+/* frontend.c: definitions of the dbus and scm-enhanced dbus
  *             frontend routines.
  *
- * Copyright © 2013 Stphane Graber
- * Author: Stphane Graber <stgraber@ubuntu.com>
+ * Copyright © 2013 Stephane Graber
+ * Author: Stephane Graber <stgraber@ubuntu.com>
  * Copyright © 2014 Canonical Group Limited
  * Author: Serge Hallyn <serge.hallyn@ubuntu.com>
  *
@@ -24,6 +24,7 @@
 #include <frontend.h>
 
 int daemonise = FALSE;
+int sigstop = FALSE;
 bool setns_pid_supported = false;
 unsigned long mypidns;
 bool setns_user_supported = false;
@@ -45,7 +46,7 @@ static struct scm_sock_data *alloc_scm_sock_data(NihDBusMessage *message,
 		int fd, enum req_type t)
 {
 	struct scm_sock_data *d;
-	int optval = -1, dbusfd;
+	int optval = 1, dbusfd;
 	socklen_t len;
 
 	if (setsockopt(fd, SOL_SOCKET, SO_PASSCRED, &optval, sizeof(optval)) == -1) {
@@ -76,27 +77,6 @@ static struct scm_sock_data *alloc_scm_sock_data(NihDBusMessage *message,
 	return d;
 }
 
-#if 0
-static const char *req_type_to_str(enum req_type r)
-{
-	switch(r) {
-		case REQ_TYPE_GET_PID: return "get_pid";
-		case REQ_TYPE_MOVE_PID: return "move_pid";
-		case REQ_TYPE_MOVE_PID_ABS: return "move_pid";
-		case REQ_TYPE_CREATE: return "create";
-		case REQ_TYPE_CHOWN: return "chown";
-		case REQ_TYPE_GET_VALUE: return "get_value";
-		case REQ_TYPE_SET_VALUE: return "set_value";
-		case REQ_TYPE_REMOVE: return "remove";
-		case REQ_TYPE_GET_TASKS: return "get_tasks";
-		case REQ_TYPE_CHMOD: return "chmod";
-		case REQ_TYPE_LIST_CHILDREN: return "list_children";
-		case REQ_TYPE_REMOVE_ON_EMPTY: return "remove_on_empty";
-		default: return "invalid";
-	}
-}
-#endif
-
 /*
  * All Scm-enhanced transactions take at least one SCM cred,
  * the requestor's.  Some require a second SCM cred to identify
@@ -106,6 +86,7 @@ static bool need_two_creds(enum req_type t)
 {
 	switch (t) {
 	case REQ_TYPE_GET_PID:
+	case REQ_TYPE_GET_PID_ABS:
 	case REQ_TYPE_MOVE_PID:
 	case REQ_TYPE_MOVE_PID_ABS:
 	case REQ_TYPE_CHOWN:
@@ -119,10 +100,6 @@ static void scm_sock_error_handler (void *data, NihIo *io)
 {
 	struct scm_sock_data *d = data;
 	NihError *error = nih_error_get ();
-#if 0
-	nih_info("got an error, type %s", req_type_to_str(d->type));
-	nih_info("error %s", strerror(error->number));
-#endif
 	nih_free(error);
 	d->fd = -1;
 }
@@ -131,10 +108,14 @@ static void scm_sock_close (struct scm_sock_data *data, NihIo *io)
 {
 	nih_assert (data);
 	nih_assert (io);
+
+	nih_free (io);
+
+	// Only delete data struct after io, because freeing io may call methods
+	// like error_handler that use this data struct.
 	if (data->fd != -1)
 		close (data->fd);
 	nih_free (data);
-	nih_free (io);
 }
 
 /*
@@ -181,6 +162,7 @@ static void sock_scm_reader(struct scm_sock_data *data,
 
 	switch (data->type) {
 	case REQ_TYPE_GET_PID: get_pid_scm_complete(data); break;
+	case REQ_TYPE_GET_PID_ABS: get_pid_abs_scm_complete(data); break;
 	case REQ_TYPE_MOVE_PID: move_pid_scm_complete(data); break;
 	case REQ_TYPE_MOVE_PID_ABS: move_pid_abs_scm_complete(data); break;
 	case REQ_TYPE_CREATE: create_scm_complete(data); break;
@@ -192,6 +174,7 @@ static void sock_scm_reader(struct scm_sock_data *data,
 	case REQ_TYPE_GET_TASKS: get_tasks_scm_complete(data); break;
 	case REQ_TYPE_LIST_CHILDREN: list_children_scm_complete(data); break;
 	case REQ_TYPE_REMOVE_ON_EMPTY: remove_on_empty_scm_complete(data); break;
+	case REQ_TYPE_PRUNE: prune_scm_complete(data); break;
 	default:
 		nih_fatal("%s: bad req_type %d", __func__, data->type);
 		exit(1);
@@ -254,11 +237,8 @@ int cgmanager_get_pid_cgroup_scm (void *data, NihDBusMessage *message,
 		return -1;
 	}
 
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -320,6 +300,137 @@ int cgmanager_get_pid_cgroup (void *data, NihDBusMessage *message,
 	return 0;
 }
 
+void get_pid_abs_scm_complete(struct scm_sock_data *data)
+{
+	// output will be nih_alloced with data as parent, and therefore
+	// freed when data is freed.
+	char *output = NULL;
+	int ret;
+
+	ret = get_pid_cgroup_abs_main(data, data->controller, data->pcred,
+			data->rcred, data->vcred, &output);
+	if (ret == 0)
+		ret = write(data->fd, output, strlen(output)+1);
+	else
+		// Let the client know it failed
+		ret = write(data->fd, &data->rcred, 0);
+	if (ret < 0)
+		nih_error("GetPidCgroupAbsScm: Error writing final result to client: %s",
+			strerror(errno));
+}
+
+/*
+ * This is one of the dbus callbacks.
+ * Caller requests the cgroup of @pid in a given @controller, relative
+ * to the proxy's
+ */
+int cgmanager_get_pid_cgroup_abs_scm (void *data, NihDBusMessage *message,
+			const char *controller, int sockfd)
+{
+	struct scm_sock_data *d;
+
+	d = alloc_scm_sock_data(message, sockfd, REQ_TYPE_GET_PID_ABS);
+	if (!d)
+		return -1;
+	d->controller = NIH_MUST( nih_strdup(d, controller) );
+
+	if (!nih_io_reopen(NULL, sockfd, NIH_IO_MESSAGE,
+				(NihIoReader) sock_scm_reader,
+				(NihIoCloseHandler) scm_sock_close,
+				scm_sock_error_handler, d)) {
+		NihError *error = nih_error_steal ();
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+			"Failed queue scm message: %s", error->message);
+		nih_free(error);
+		return -1;
+	}
+
+	if (!kick_fd_client(sockfd))
+		return -1;
+	return 0;
+}
+
+/* GetPidCgroup */
+/*
+ * This is one of the dbus callbacks.
+ * Caller requests the cgroup of @pid in a given @controller relative
+ * to the proxy's
+ */
+int cgmanager_get_pid_cgroup_abs (void *data, NihDBusMessage *message,
+			const char *controller, int plain_pid, char **output)
+{
+	int fd = 0, ret;
+	struct ucred rcred, vcred;
+	socklen_t len;
+
+	if (message == NULL) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+			"message was null");
+		return -1;
+	}
+
+	if (!dbus_connection_get_socket(message->connection, &fd)) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "Could not get client socket.");
+		return -1;
+	}
+
+	len = sizeof(struct ucred);
+	if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &rcred, &len) < 0) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "Could not get peer cred: %s",
+					     strerror(errno));
+		return -1;
+	}
+
+	nih_info (_("GetPidCgroupAbs: Client fd is: %d (pid=%d, uid=%u, gid=%u)"),
+			fd, rcred.pid, rcred.uid, rcred.gid);
+
+	/*
+	 * getpidcgroup results cannot make sense as the pid is not
+	 * translated.  Note that on an old enough kernel we cannot detect
+	 * this situation.  In that case we allow it - it will confuse the
+	 * caller, but cause no harm
+	 */
+	if (!is_same_pidns(rcred.pid)) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+				"GetPidCgroupAbs called from non-init namespace");
+		return -1;
+	}
+	vcred.uid = 0;
+	vcred.gid = 0;
+	vcred.pid = plain_pid;
+
+#ifdef CGMANAGER
+	/*
+	 * A plain dbus request to escape cgroup root was made by a root
+	 * owned task in cgmanager's namespace.  We will send ourselves as the
+	 * proxy.
+	 */
+	struct ucred mycred = {
+		.pid = getpid(),
+		.uid = getuid(),
+		.gid = getgid()
+	};
+#else
+	/*
+	 * This is the !CGMANAGER case.  We are in the proxy.  We don't
+	 * support chained proxying anyway, so it is simple - the requestor
+	 * is the proxy at this point;  then we will proxy the call on to
+	 * the cgmanager
+	 */
+#define mycred rcred
+#endif
+
+	ret = get_pid_cgroup_abs_main(message, controller, mycred, rcred, vcred, output);
+	if (ret) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+				"invalid request");
+		return -1;
+	}
+	return 0;
+}
+
 void move_pid_scm_complete(struct scm_sock_data *data)
 {
 	char b = '0';
@@ -353,11 +464,8 @@ int cgmanager_move_pid_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -446,11 +554,8 @@ int cgmanager_move_pid_abs_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -566,11 +671,8 @@ int cgmanager_create_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -651,11 +753,8 @@ int cgmanager_chown_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -758,11 +857,8 @@ int cgmanager_chmod_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -845,11 +941,8 @@ int cgmanager_get_value_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 
 }
@@ -937,11 +1030,8 @@ int cgmanager_set_value_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -1027,11 +1117,8 @@ int cgmanager_remove_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -1079,6 +1166,7 @@ int cgmanager_remove (void *data, NihDBusMessage *message, const char *controlle
 	return ret;
 }
 
+/* get_tasks - list tasks for a single cgroup */
 void get_tasks_scm_complete(struct scm_sock_data *data)
 {
 	struct ucred pcred;
@@ -1127,11 +1215,8 @@ int cgmanager_get_tasks_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -1182,6 +1267,111 @@ int cgmanager_get_tasks (void *data, NihDBusMessage *message, const char *contro
 	return ret;
 }
 
+/* GetTasksRecursive - list tasks for a cgroup and any descendents
+ * inherintly racy. */
+void get_tasks_recursive_scm_complete(struct scm_sock_data *data)
+{
+	struct ucred pcred;
+	int i, ret;
+	int32_t *pids, nrpids;
+	ret = get_tasks_main(data, data->controller, data->cgroup,
+			data->pcred, data->rcred, &pids);
+	if (ret < 0) {
+		nih_error("Error getting nrtasks for %s:%s for pid %d",
+			data->controller, data->cgroup, data->rcred.pid);
+		ret = -1;
+	}
+	nrpids = ret;
+	if (write(data->fd, &nrpids, sizeof(int32_t)) != sizeof(int32_t)) {
+		nih_error("get_tasks_recursive_scm: Error writing final result to client");
+		return;
+	}
+	pcred.uid = 0; pcred.gid = 0;
+	for (i=0; i<ret; i++) {
+		pcred.pid = pids[i];
+		if (send_creds(data->fd, &pcred)) {
+			nih_error("get_tasks_recursive_scm: error writing pids back to client");
+			return;
+		}
+	}
+}
+
+int cgmanager_get_tasks_recursive_scm (void *data, NihDBusMessage *message,
+		 const char *controller, const char *cgroup, int sockfd)
+{
+	struct scm_sock_data *d;
+
+	d = alloc_scm_sock_data(message, sockfd, REQ_TYPE_GET_TASKS_RECURSIVE);
+	if (!d)
+		return -1;
+	d->controller = NIH_MUST( nih_strdup(d, controller) );
+	d->cgroup = NIH_MUST( nih_strdup(d, cgroup) );
+
+	if (!nih_io_reopen(NULL, sockfd, NIH_IO_MESSAGE,
+				(NihIoReader) sock_scm_reader,
+				(NihIoCloseHandler) scm_sock_close,
+				scm_sock_error_handler, d)) {
+		NihError *error = nih_error_steal ();
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+			"Failed queue scm message: %s", error->message);
+		nih_free(error);
+		return -1;
+	}
+	if (!kick_fd_client(sockfd))
+		return -1;
+	return 0;
+}
+
+/* 
+ * This is one of the dbus callbacks.
+ * Caller requests the number of tasks in @cgroup in @controller
+ * returns nrpids, or -1 on error.
+ */
+int cgmanager_get_tasks_recursive (void *data, NihDBusMessage *message,
+		const char *controller, const char *cgroup, int32_t **pids,
+		size_t *nrpids)
+{
+	int fd = 0, ret;
+	struct ucred rcred;
+	socklen_t len;
+	int32_t *tmp;
+
+	if (message == NULL) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+			"message was null");
+		return -1;
+	}
+
+	if (!dbus_connection_get_socket(message->connection, &fd)) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "Could not get client socket.");
+		return -1;
+	}
+
+	len = sizeof(struct ucred);
+	if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &rcred, &len) < 0) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "Could not get peer cred: %s",
+					     strerror(errno));
+		return -1;
+	}
+
+	nih_info (_("GetTasksRecursive: Client fd is: %d (pid=%d, uid=%u, gid=%u)"),
+			fd, rcred.pid, rcred.uid, rcred.gid);
+
+	ret = get_tasks_recursive_main(message, controller, cgroup, rcred, rcred, &tmp);
+	if (ret >= 0) {
+		*nrpids = ret;
+		*pids = tmp;
+		ret = 0;
+	} else
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "invalid request");
+	return ret;
+}
+
+
+/* ListChildren - list child cgroups */
 void list_children_scm_complete(struct scm_sock_data *data)
 {
 	int i, ret;
@@ -1254,11 +1444,8 @@ int cgmanager_list_children_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -1340,11 +1527,8 @@ int cgmanager_remove_on_empty_scm (void *data, NihDBusMessage *message,
 		nih_free(error);
 		return -1;
 	}
-	if (!kick_fd_client(sockfd)) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-			"Error writing to client: %s", strerror(errno));
+	if (!kick_fd_client(sockfd))
 		return -1;
-	}
 	return 0;
 }
 
@@ -1393,6 +1577,139 @@ int cgmanager_remove_on_empty (void *data, NihDBusMessage *message,
 	return ret;
 }
 
+/*
+ * Prune - recursively call remove-on-empty first, then remove, on each
+ * directory.
+ */
+void prune_scm_complete(struct scm_sock_data *data)
+{
+	char b = '0';
+
+	if (prune_main(data->controller, data->cgroup, data->pcred,
+				data->rcred) == 0)
+		b = '1';
+	if (write(data->fd, &b, 1) < 0)
+		nih_error("PruneScm: Error writing final result to client");
+}
+
+int cgmanager_prune_scm (void *data, NihDBusMessage *message,
+		 const char *controller, const char *cgroup, int sockfd)
+{
+	struct scm_sock_data *d;
+
+	d = alloc_scm_sock_data(message, sockfd, REQ_TYPE_PRUNE);
+	if (!d)
+		return -1;
+	d->controller = NIH_MUST( nih_strdup(d, controller) );
+	d->cgroup = NIH_MUST( nih_strdup(d, cgroup) );
+
+	if (!nih_io_reopen(NULL, sockfd, NIH_IO_MESSAGE,
+				(NihIoReader) sock_scm_reader,
+				(NihIoCloseHandler) scm_sock_close,
+				scm_sock_error_handler, d)) {
+		NihError *error = nih_error_steal ();
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+			"Failed queue scm message: %s", error->message);
+		nih_free(error);
+		return -1;
+	}
+	if (!kick_fd_client(sockfd))
+		return -1;
+	return 0;
+}
+
+/* 
+ * This is one of the dbus callbacks.
+ * Caller requests that cgroup @cgroup in controller @controller be
+ * recursively removed if empty, or else removed when it becomes empty.
+ */
+int cgmanager_prune (void *data, NihDBusMessage *message,
+		const char *controller, const char *cgroup)
+{
+	int fd = 0, ret;
+	struct ucred rcred;
+	socklen_t len;
+
+	if (message == NULL) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+			"message was null");
+		return -1;
+	}
+
+	if (!dbus_connection_get_socket(message->connection, &fd)) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "Could not get client socket.");
+		return -1;
+	}
+
+	len = sizeof(struct ucred);
+	if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &rcred, &len) < 0) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "Could not get peer cred: %s",
+					     strerror(errno));
+		return -1;
+	}
+
+	nih_info (_("Prune: Client fd is: %d (pid=%d, uid=%u, gid=%u)"),
+			fd, rcred.pid, rcred.uid, rcred.gid);
+
+	ret = prune_main(controller, cgroup, rcred, rcred);
+	if (ret >= 0)
+		ret = 0;
+	else
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "invalid request");
+	return ret;
+}
+
+/*
+ * listcontrollers
+ */
+int cgmanager_list_controllers (void *data, NihDBusMessage *message,
+		char ***output)
+{
+	int fd = 0, ret;
+	struct ucred rcred;
+	socklen_t len;
+
+	nih_assert(output);
+
+	if (message == NULL) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+			"message was null");
+		return -1;
+	}
+
+	if (!dbus_connection_get_socket(message->connection, &fd)) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "Could not get client socket.");
+		return -1;
+	}
+
+	len = sizeof(struct ucred);
+	if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &rcred, &len) < 0) {
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "Could not get peer cred: %s",
+					     strerror(errno));
+		return -1;
+	}
+
+	nih_info (_("ListControllers: Client fd is: %d (pid=%d, uid=%u, gid=%u)"),
+			fd, rcred.pid, rcred.uid, rcred.gid);
+
+	ret = list_controllers_main(message, output);
+	if (ret >= 0)
+		ret = 0;
+	else
+		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
+					     "invalid request");
+	return ret;
+}
+
+
+/*
+ * return our APi version
+ */
 int
 cgmanager_get_api_version(void *data, NihDBusMessage *message, int *version)
 {
